#include <cstring>
#include <sys/ioctl.h>

#include "AlsaClient.h"
#include "util/schedYield.h"
#include "util/mytype.h"


AlsaClient::AlsaClient (int mode)
{
  int ret = snd_seq_open (&handle, mode);
  if (ret < 0)
    {
      cerr << "AlsaClient::AlsaClient " << snd_strerror (ret) << endl;
      throw CanNotOpenDevice ();
    }
  clientNumber = snd_seq_client_id (handle);

  //NO NEED this check?
  if (clientNumber < 0)
    {
      snd_seq_close (handle);
      throw CanNotGetClientID ();
    }
}

AlsaClient::~AlsaClient ()
{
  int ret = snd_seq_close (handle);
  if (ret < 0)
    cerr << "Alsa::~Alsa:can not close device:" << snd_strerror (ret) << endl;
}

////////////////////////////////////////////////////////////system info
void AlsaClient::getPVersion (int* version)
{
  assert (version != NULL);
  const int fd = snd_seq_file_descriptor (handle);
  int ret = ioctl (fd, SND_SEQ_IOCTL_PVERSION, version);
  checkResult (ret);
}

void AlsaClient::getSystemInfo(snd_seq_system_info_t * info)
{
  int ret = snd_seq_system_info (handle, info);
  checkResult (ret);
}


////////////////////////////////////////////////////////////client
int AlsaClient::blockMode (bool enable)
{
  int ret = snd_seq_block_mode (handle, enable);
  checkResult (ret);
  return ret;
}

  
#if 0
////////////////////////////////////////////////////////////event
////////////////////////////////////////////////////////////memory
static snd_seq_event_t* newEvent ()
{
}

static void deleteEvent (snd_seq_event_t* ev)
{
}

static int eventLength (snd_seq_event_t& event)
{
}

#endif
  
////////////////////////////////////////////////////////////queue
int AlsaClient::useQueue (int queue)
{
  return setQueueInfo (queue, false, false, true);
}


int AlsaClient::ownQueue (int queue)
{
  return setQueueInfo (queue, true, false, true);
}

int AlsaClient::lockQueue (int queue)
{
  return setQueueInfo (queue, true, true, true);
}

int AlsaClient::freeQueue (int queue)
{
  return setQueueInfo (queue, false, false, false);
}

  
int AlsaClient::setQueueClientInfo (snd_seq_queue_client_t* info)
{
  return snd_seq_set_queue_client (handle, info->queue, info);
}

int AlsaClient::setQueueClientInfo (int queue, bool used,
				    int low = -1/*unchange*/,
				    int high = -1/*unchange*/)
{
  snd_seq_queue_client_t info;
  int ret = snd_seq_get_queue_client (handle, queue, &info);
  if (ret < 0)
    return ret;
  info.used = used;
  if (low != -1)
    info.low = low;
  if (high != -1)
    info.high = high;
  return snd_seq_set_queue_client (handle, queue, &info);
}

int AlsaClient::getQueueClientInfo (int queue, snd_seq_queue_client_t* info)
{
  return snd_seq_get_queue_client (handle, queue, info);
}

int AlsaClient::getQueueClientInfo (snd_seq_queue_client_t* info)
{
  return snd_seq_get_queue_client (handle, info->queue, info);
}

  
int AlsaClient::getQueueInfo (snd_seq_queue_info_t* info)
{
  return snd_seq_get_queue_info (handle, info->queue, info);
}

int AlsaClient::setQueueInfo (int queue, bool own, bool lock, bool used)
{
  snd_seq_queue_info_t info;
  int ret = snd_seq_get_queue_info (handle, queue, &info);
  if (ret < 0)
    return ret;
  info.owner = (own? getClientId (): -1);
  info.used = (used? 1: 0);
  info.locked = (lock? 1: 0);
  return snd_seq_set_queue_info (handle, queue, &info);
}


int AlsaClient::setQueueInfo (int queue, int tempo, int ppq)
{
  snd_seq_queue_info_t info;
  int ret = snd_seq_get_queue_info (handle, queue, &info);
  if (ret < 0)
    return ret;
  info.tempo = tempo;
  info.ppq = ppq;
  return snd_seq_set_queue_info (handle, queue, &info);
}

int AlsaClient::setQueueInfo (snd_seq_queue_info_t* info)
{
  return snd_seq_set_queue_info (handle, info->queue, info);
}


  
  ////////////////////////////////////////////////////////////subscribe
int AlsaClient::outputSubscribe (int port, int queue,
				 int destClient, int destPort,
				 bool exclusive = false, bool realtime = false)
{
  snd_seq_port_subscribe_t subscribe;
  subscribe.sender.queue = queue;
  subscribe.sender.client = getClientId ();
  subscribe.sender.port = port;
  subscribe.sender.channel = 0; //don't care
  subscribe.dest.queue = queue;
  subscribe.dest.client = destClient;
  subscribe.dest.port = destPort;
  subscribe.dest.channel = 0;//don't care
  subscribe.exclusive = exclusive;
  subscribe.realtime = realtime;

  return snd_seq_subscribe_port (handle, &subscribe);
}

int AlsaClient::outputUnsubscribe (int port, int queue,
				   int destClient, int destPort)
{
  snd_seq_port_subscribe_t subscribe;
  subscribe.sender.queue = queue;
  subscribe.sender.client = getClientId ();
  subscribe.sender.port = port;
  subscribe.sender.channel = 0; //don't care
  subscribe.dest.queue = queue;
  subscribe.dest.client = destClient;
  subscribe.dest.port = destPort;
  subscribe.dest.channel = 0;//don't care
  subscribe.exclusive = 0;//don't care
  subscribe.realtime = 0;//don't care

  return snd_seq_unsubscribe_port (handle, &subscribe);
}


int AlsaClient::inputSubscribe (int srcClient, int srcPort,
				int queue, int port,
				bool exclusive = false, bool realtime = false)
{
  snd_seq_port_subscribe_t subscribe;
  subscribe.sender.queue = queue;
  subscribe.sender.client = srcClient;
  subscribe.sender.port = srcPort;
  subscribe.sender.channel = 0; //don't care
  subscribe.dest.queue = queue;
  subscribe.dest.client = getClientId ();
  subscribe.dest.port = port;
  subscribe.dest.channel = 0;//don't care
  subscribe.exclusive = exclusive;
  subscribe.realtime = realtime;

  return snd_seq_subscribe_port (handle, &subscribe);
}

int AlsaClient::inputUnsubscribe (int srcClient, int srcPort, int queue,
				  int port)
{
  snd_seq_port_subscribe_t subscribe;
  subscribe.sender.queue = queue;
  subscribe.sender.client = srcClient;
  subscribe.sender.port = srcPort;
  subscribe.sender.channel = 0; //don't care
  subscribe.dest.queue = queue;
  subscribe.dest.client = getClientId ();
  subscribe.dest.port = port;
  subscribe.dest.channel = 0;//don't care
  subscribe.exclusive = 0;//don't care
  subscribe.realtime = 0;//don't care

  return snd_seq_subscribe_port (handle, &subscribe);  
}


////////////////////////////////////////////////////////////client info
int AlsaClient::getClientInfo (snd_seq_client_info_t* clientInfo)
{
  return snd_seq_get_client_info (handle, clientInfo);
}

int AlsaClient::getClientInfo (int client, snd_seq_client_info_t* clientInfo)
{
  return snd_seq_get_any_client_info (handle, client, clientInfo);
}

int AlsaClient::setClientInfo (//64 is from snd_seq_client_info_t
			       char name[64],
			       unsigned int filter,
			       //8 is from snd_seq_client_info_t
			       unsigned char multicastFilter[8],
			       //32 is from snd_seq_client_info_t
			       unsigned char eventFilter[32])
{
  assert (name != NULL);
  assert (multicastFilter != NULL);
  assert (eventFilter != NULL);
  
  snd_seq_client_info_t info;
  info.client = getClientId ();
  info.type = USER_CLIENT;
  memcpy (info.name, name, 64);
  info.filter = filter;
  memcpy (info.multicast_filter, multicastFilter, 8);
  memcpy (info.event_filter, eventFilter, 32);
  return snd_seq_set_client_info (handle, &info);
}



////////////////////////////////////////////////////////////port
//portֹϼưŪ˳꿶롣
//portξportInfo㤨롣
int AlsaClient::createPort (snd_seq_port_info_t* portInfo)
{
  return snd_seq_create_port (handle, portInfo);
}

int AlsaClient::createPort (const char* name, int capability, int type,
			    int midiChannels, int synthVoices, int use,
			    snd_seq_port_info_t* portInfo)
{
  assert (portInfo != NULL);
  strncpy (portInfo->name, name, sizeof (portInfo->name));
  portInfo->capability = capability;
  portInfo->type = type;
  portInfo->midi_channels = midiChannels;
  portInfo->synth_voices = synthVoices;
  portInfo->use = use;
  portInfo->kernel = NULL;

  return snd_seq_create_port (handle, portInfo);
}


int AlsaClient::deletePort (int port)
{
  snd_seq_port_info_t portInfo;
  //??? Is this enough?
  portInfo.port = port;
  return snd_seq_delete_port (handle, &portInfo);
}

int AlsaClient::getPortInfo (int client, int port,
			     snd_seq_port_info_t* portInfo)
{
  return snd_seq_get_any_port_info (handle, client, port, portInfo);
}

int AlsaClient::getPortInfo (int port, snd_seq_port_info_t* portInfo)
{
  return snd_seq_get_port_info (handle, port, portInfo);
}

int AlsaClient::setPortInfo (int port,
			     //64 is from snd_seq_port_info_t
			     char name[64],
			     unsigned int capability, unsigned int type,
			     int midiChannels, int synthVoices)
{
  assert (name != NULL);
  
  snd_seq_port_info_t portInfo;
  portInfo.client = getClientId ();
  portInfo.port = port;
  memcpy (portInfo.name, name, 64);
  portInfo.capability = capability;
  portInfo.type = type;
  portInfo.midi_channels = midiChannels;
  portInfo.synth_voices = synthVoices;
  return snd_seq_set_port_info (handle, port, &portInfo);
}

  
  
////////////////////////////////////////////////write event
int AlsaClient::eventOutput (snd_seq_event_t* event)
{
  assert (event != NULL);

  //XXXX
#if 0
  const int COUNT_MAX = 500;
  const int COUNT_MIN = 50;
  static int count = 0;
  count++;
  if (count >= COUNT_MAX)
    {
      while (snd_seq_flush_output(handle) > COUNT_MIN)
	sched_yield ();
      count = 0;
    }
#endif

#if 1
  int ret = -ENOMEM;
  while (ret < 0)
    {
      ret = snd_seq_event_output (handle, event);
      if (ret < 0)
	{
	  cerr << "AlsaClient::eventOutput ret = " << ret
	       << " (" << snd_strerror (ret) << ")" << endl;
	  //sleep(1);
	  schedYield ();
	}
    }
  return ret;
#else
  return snd_seq_event_output (handle, event);
#endif
}


#if 0
//???
//SND_SEQ_EVENT_SYSTEM
int AlsaClient::systemEvent ()
{
  
}
#endif
  
//SND_SEQ_EVENT_NOTE
inline int AlsaClient::doNote (snd_seq_event_t* event,
			       unsigned char note, unsigned char velocity,
			       unsigned int duration)
{
  event->type = SND_SEQ_EVENT_NOTE;
  event->flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event->data.note.note = note;
  event->data.note.velocity = velocity;
  event->data.note.duration = duration;

  return eventOutput (event);
}


int AlsaClient::note (unsigned char note, unsigned char velocity,
		      unsigned int duration,
		      snd_seq_tick_time_t ticks,
		      int srcPort, int queue,
		      int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);
  return doNote (&event, note, velocity, duration);
}

int AlsaClient::note (unsigned char note, unsigned char velocity,
		      unsigned int duration,
		      snd_seq_real_time_t real,
		      int srcPort, int queue,
		      int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);
  return doNote (&event, note, velocity, duration);
}

//SND_SEQ_EVENT_NOTEON
inline int AlsaClient::doNoteOn (snd_seq_event_t* event,
				 unsigned char note, unsigned char velocity)
{
  event->type = SND_SEQ_EVENT_NOTEON;
  event->flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event->data.note.note = note;
  event->data.note.velocity = velocity;
  //event.data.note.duration;

  return eventOutput (event);
}


int AlsaClient::noteOn (unsigned char note, unsigned char velocity,
			snd_seq_tick_time_t ticks,
			int srcPort, int queue,
			int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);

  return doNoteOn (&event, note, velocity);
}

int AlsaClient::noteOn (unsigned char note, unsigned char velocity,
			snd_seq_real_time_t real,
			int srcPort, int queue,
			int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);

  return doNoteOn (&event, note, velocity);
}

//SND_SEQ_EVENT_NOTEOFF
inline int AlsaClient::doNoteOff (snd_seq_event_t* event,
				  unsigned char note, unsigned char velocity)
{
  event->type = SND_SEQ_EVENT_NOTEOFF;
  event->flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event->data.note.note = note;
  event->data.note.velocity = velocity;
  //event.data.note.duration;

  return eventOutput (event);
}


int AlsaClient::noteOff (unsigned char note, unsigned char velocity,
			 snd_seq_tick_time_t ticks,
			 int srcPort, int queue,
			 int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);
  return doNoteOff (&event, note, velocity);
}

int AlsaClient::noteOff (unsigned char note, unsigned char velocity,
			 snd_seq_real_time_t real,
			 int srcPort, int queue,
			 int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);
  return doNoteOff (&event, note, velocity);
}

//SND_SEQ_EVENT_KEYPRESS
inline int AlsaClient::doKeyPressure (snd_seq_event_t* event,
				      unsigned char note,
				      unsigned char keyPressure)
{
  event->type = SND_SEQ_EVENT_KEYPRESS;
  event->flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event->data.control.param = note;
  event->data.control.value = keyPressure;

  return eventOutput (event);
}

int AlsaClient::keyPressure (unsigned char note, unsigned char keyPressure,
			     snd_seq_tick_time_t ticks,
			     int srcPort, int queue,
			     int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);

  return doKeyPressure (&event, note, keyPressure);
}

int AlsaClient::keyPressure (unsigned char note, unsigned char keyPressure,
			     snd_seq_real_time_t real,
			     int srcPort, int queue,
			     int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);

  return doKeyPressure (&event, note, keyPressure);
}


//controlChange
//SND_SEQ_EVENT_CONTROLLER
int AlsaClient::doController (snd_seq_event_t* event,
			      unsigned char controll, unsigned char value)
{
  event->type = SND_SEQ_EVENT_CONTROLLER;
  event->flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event->data.control.param = controll;
  event->data.control.value = value;

  return eventOutput (event);
}


int AlsaClient::controller (unsigned char controll, unsigned char value,
			    snd_seq_tick_time_t ticks,
			    int srcPort, int queue,
			    int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);
  return doController (&event, controll, value);
}

int AlsaClient::controller (unsigned char controll, unsigned char value,
			    snd_seq_real_time_t real,
			    int srcPort, int queue,
			    int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);
  return doController (&event, controll, value);
}

//SND_SEQ_EVENT_PGMCHANGE
int AlsaClient::doProgramChange (snd_seq_event_t* event,
				 unsigned char program)
{
  event->type = SND_SEQ_EVENT_PGMCHANGE;
  event->flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event->data.control.value = program;

  return eventOutput (event);
}

int AlsaClient::programChange (unsigned char program,
			       snd_seq_tick_time_t ticks,
			       int srcPort, int queue,
			       int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);
  
  return doProgramChange (&event, program);
}

int AlsaClient::programChange (unsigned char program,
			       snd_seq_real_time_t real,
			       int srcPort, int queue,
			       int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);
  
  return doProgramChange (&event, program);
}

//SND_SEQ_EVENT_CHANPRESS
int AlsaClient::doChannelPressure (snd_seq_event_t* event,
				   unsigned char channelPressure)
{
  event->type = SND_SEQ_EVENT_CHANPRESS;
  event->flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event->data.control.value = channelPressure;

  return eventOutput (event);
}


int AlsaClient::channelPressure (unsigned char channelPressure,
				 snd_seq_tick_time_t ticks,
				 int srcPort, int queue,
				 int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);
  return doChannelPressure (&event, channelPressure);
}

int AlsaClient::channelPressure (unsigned char channelPressure,
				 snd_seq_real_time_t real,
				 int srcPort, int queue,
				 int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);
  return doChannelPressure (&event, channelPressure);
}

//SND_SEQ_EVENT_PITCHBEND
int AlsaClient::doPitchBend (snd_seq_event_t* event,
			     unsigned char lsb, unsigned char msb)
{
  event->type = SND_SEQ_EVENT_PITCHBEND;
  event->flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event->data.control.value = (lsb + (msb << 7)) - 8192;

  return eventOutput (event);
}


int AlsaClient::pitchBend (unsigned char lsb, unsigned char msb,
			   snd_seq_tick_time_t ticks,
			   int srcPort, int queue,
			   int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);

  return doPitchBend (&event, lsb, msb);
}

int AlsaClient::pitchBend (unsigned char lsb, unsigned char msb,
			   snd_seq_real_time_t real,
			   int srcPort, int queue,
			   int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);

  return doPitchBend (&event, lsb, msb);
}

#if 0
//???
//SND_SEQ_EVENT_CONTROL14
int AlsaClient::control14 (snd_seq_tick_time_t ticks,
			   int srcPort, int queue, int destClient, int destPort, int channel)
{
}

#endif
#if 0
//???
//SND_SEQ_EVENT_NONREGPARAM
int AlsaClient::nonRegProg (snd_seq_tick_time_t ticks,
			    int srcPort, int queue,
			    int destClient, int destPort, int channel)
{
}

  
//???
//SND_SEQ_EVENT_REGPARAM
int AlsaClient::regProg (snd_seq_tick_time_t ticks,
			 int srcPort, int queue,
			 int destClient, int destPort, int channel)
{
}

#endif

#if 0
//SND_SEQ_EVENT_SONGPOS/* Song Position Pointer with LSB and MSB values */
//0xF2 lsb msb
int AlsaClient::songPosition (unsigned char lsb, unsgined char msb,
			      snd_seq_tick_time_t ticks,
			      int srcPort, int queue,
			      int destClient, int destPort, int channel)
{
}

//SND_SEQ_EVENT_SONGSEL		/* Song Select with song ID number */
//0xF3 dd
int AlsaClient::songSelct (unsgined char songNumber,
			   snd_seq_tick_time_t ticks,
			   int srcPort, int queue,
			   int destClient,ort, int channel)
{
}

//SND_SEQ_EVENT_CLOCK		 midi Real Time Clock message */
//0xF8
int AlsaClient::timingClock (snd_seq_tick_time_t ticks,
			     int srcPort, int queue,
			     int destClient,ort, int channel)
{
}

//SND_SEQ_EVENT_START		 midi Real Time Start message */
//0xFA
int AlsaClient::start (snd_seq_tick_time_t ticks,
		       int srcPort, int queue, int destClient,ort, int channel)
{
  snd_seq_event_t event;
  snd_seq_tick_real_time_t realTime;
  realTime.tv_sec = 0;
  realTime.tv_nsec = 0;
  setOutputEventHeader (&event, realTime, SND_SEQ_PORT_SYSTEM_TIMER,
			queue, SND_SEQ_CLIENT_SYSTEM, 0, 255);
  event.flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event.type = SND_SEQ_EVENT_START;
  
  eventOutput (&event);
}

//SND_SEQ_EVENT_CONTINUE		/* midi Real Time Continue message */
//0xFB
int AlsaClient::eventContinue (snd_seq_tick_time_t ticks,
			       int srcPort, int queue,
			       int destClient,ort, int channel)
{
}

//SND_SEQ_EVENT_STOP		 midi Real Time Stop message */
//0xFC
int AlsaClient::eventStop (snd_seq_tick_time_t ticks,
			   int srcPort, int queue,
			   int destClient,ort, int channel)
{
}

//SND_SEQ_EVENT_QFRAME		/* Midi time code quarter frame */
//0xF1 mX
int AlsaClient::qframe (unsgined char messageType, unsgined char value,
			snd_seq_tick_time_t ticks,
			int srcPort, int queue, int destClient,ort, int channel)
{
}

int AlsaClient::qframe (unsgined char frameCount,
			unsigned char second, unsigned char minuite,
			unsigned char hour,
			snd_seq_tick_time_t ticks,
			int srcPort, int queue, int destClient,ort, int channel)
{
}

#endif
  
#if 0
//SND_SEQ_EVENT_TEMPO   /* (SMF) Tempo event */
int AlsaClient::tempo (int tempo, snd_seq_tick_time_t ticks,
		       int srcPort, int queue, int destClient, int destPort,
		       int channel = 0/*don't care*/)
{
}

//SND_SEQ_EVENT_TIMESIGN/* SMF Time Signature event */
int AlsaClient::timeSignature (snd_seq_tick_time_t ticks,
			       int srcPort, int queue,
			       int destClient, int destPort, int channel)
{
}

//SND_SEQ_EVENT_KEYSIGN /* SMF Key Signature event */
int AlsaClient::keySignature (snd_seq_tick_time_t ticks,
			      int srcPort, int queue,
			      int destClient, int destPort, int channel)
{
}

#endif
//SND_SEQ_EVENT_SYSEX   /* system exclusive data (variable length) */
int AlsaClient::doSysEx (snd_seq_event_t* event,
			 int length, const char* buff)
{
  event->type = SND_SEQ_EVENT_SYSEX;
  event->flags |= SND_SEQ_EVENT_LENGTH_VARIABLE;
  event->data.ext.len = length;
  event->data.ext.ptr = buff;//XXX discards const

  return eventOutput (event);
}


int AlsaClient::sysEx (int length, const char* buff,
		       snd_seq_tick_time_t ticks,
		       int srcPort, int queue,
		       int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);

  return doSysEx (&event, length, buff);
}

int AlsaClient::sysEx (int length, const char* buff,
		       snd_seq_real_time_t real,
		       int srcPort, int queue,
		       int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);

  return doSysEx (&event, length, buff);
}


#if 0
//SND_SEQ_EVENT_HEARTBEAT/* "active sensing" event */
//0xFE
int AlsaClient::heartBeat (snd_seq_tick_time_t ticks,
			   int srcPort, int queue,
			   int destClient, int destPort, int channel)
{
}
#endif

int AlsaClient::doEcho (snd_seq_event_t* event)
{
  event->type = SND_SEQ_EVENT_ECHO;
  event->flags |= SND_SEQ_EVENT_LENGTH_VARIABLE;

  return eventOutput (event);
}

//SND_SEQ_EVENT_ECHO/* echo event */
int AlsaClient::echoEvent (snd_seq_tick_time_t ticks,
			   int srcPort, int queue,
			   int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, srcPort, queue,
			destClient, destPort, channel);

  return doEcho (&event);
}

//SND_SEQ_EVENT_ECHO/* echo event */
int AlsaClient::echoEvent (snd_seq_real_time_t real,
			   int srcPort, int queue,
			   int destClient, int destPort, int channel)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, srcPort, queue,
			destClient, destPort, channel);

  return doEcho (&event);
}




#if 0
/* system status messages */
//SND_SEQ_EVENT_CLIENT_START  /* new client has connected */
int AlsaClient::clientStart (snd_seq_tick_time_t ticks,
			     int srcPort, int queue,
			     int destClient, int destPort, int channel)
{
}

//SND_SEQ_EVENT_CLIENT_EXIT   /* client has left the system */
int AlsaClient::clientExit (snd_seq_tick_time_t ticks,
			    int srcPort, int queue,
			    int destClient, int destPort, int channel)
{
}

//SND_SEQ_EVENT_CLIENT_CHANGE /* client status/info has changed */
int AlsaClient::clientChange (snd_seq_tick_time_t ticks,
			      int srcPort, int queue,
			      int destClient, int destPort, int channel)
{
}


//SND_SEQ_EVENT_PORT_START    /* new port was created */
int AlsaClient::portStart (snd_seq_tick_time_t ticks,
			   int srcPort, int queue,
			   int destClient, int destPort, int channel)
{
}

//SND_SEQ_EVENT_PORT_EXIT	    /* port was deleted from system */
int AlsaClient::portExit (snd_seq_tick_time_t ticks,
			  int srcPort, int queue,
			  int destClient, int destPort, int channel)
{
}

//SND_SEQ_EVENT_PORT_CHANGE   /* port status/info has changed */
//???
int AlsaClient::portChange (snd_seq_tick_time_t ticks,
			    int srcPort, int queue,
			    int destClient, int destPort, int channel)
{
}

#endif

#if 0
/* synthesizer events */	
//???
//SND_SEQ_EVENT_SAMPLE	     /* sample select */
int AlsaClient::sampleEvent (snd_seq_tick_time_t ticks,
			     int srcPort, int queue,
			     int destClient, int destPort, int channel)
{
}

//???
//SND_SEQ_EVENT_SAMPLE_START   /* voice start */
int AlsaClient::sampleStart (snd_seq_tick_time_t ticks,
			     int srcPort, int queue,
			     int destClient, int destPort, int channel)
{
}

//???
//SND_SEQ_EVENT_SAMPLE_STOP    /* voice stop */
int AlsaClient::sampleStop (snd_seq_tick_time_t ticks,
			    int srcPort, int queue,
			    int destClient, int destPort, int channel)
{
}

//???
//SND_SEQ_EVENT_SAMPLE_FREQ    /* playback frequency */
int AlsaClient::sampleFreq (snd_seq_tick_time_t ticks,
			    int srcPort, int queue,
			    int destClient, int destPort, int channel)
{
}

//???
//SND_SEQ_EVENT_SAMPLE_VOLUME  /* volume and balance */
int AlsaClient::sampleVolume (snd_seq_tick_time_t ticks,
			      int srcPort, int queue,
			      int destClient, int destPort, int channel)
{
}

//???
//SND_SEQ_EVENT_SAMPLE_LOOP    /* sample loop */
int AlsaClient::sampleLoop (snd_seq_tick_time_t ticks,
			    int srcPort, int queue,
			    int destClient, int destPort, int channel)
{
}

//???
//SND_SEQ_EVENT_SAMPLE_POSITION/* sample position */
int AlsaClient::samplePosition (snd_seq_tick_time_t ticks,
				int srcPort, int queue,
				int destClient, int destPort, int channel)
{
}

#endif

int AlsaClient::flushOutput ()
{
#if 0
  return snd_seq_flush_output (handle);
#else
  int left = snd_seq_flush_output (handle);
  while (left > 0)
    {
      schedYield ();
      left = snd_seq_flush_output (handle);
      if (left < 0)
	return left;
    }
  return 0;
#endif
}

int AlsaClient::drainOutput ()
{
  return snd_seq_drain_output (handle);
}


////////////////////////////////////////////////////////////timer

void AlsaClient::startTimer (int destQueue)
{
  DEBUG_MESSAGE ("AlsaClient::startTiemr called");
  snd_seq_event_t event;
  snd_seq_real_time_t realTime;
  realTime.tv_sec = 0;
  realTime.tv_nsec = 0;
  setOutputEventHeader (&event, realTime, SND_SEQ_CLIENT_SYSTEM,
			destQueue, 255, 0, 0);
  event.flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event.type = SND_SEQ_EVENT_START;
  
  eventOutput (&event);
}

void AlsaClient::stopTimer (int destQueue, snd_seq_tick_time_t ticks)
{
  DEBUG_MESSAGE ("AlsaClient::stopTiemr called");
  snd_seq_event_t event;
  setOutputEventHeader (&event, ticks, 0, destQueue, 255, 0, 0);
  event.flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event.type = SND_SEQ_EVENT_STOP;
  
  eventOutput (&event);
}

void AlsaClient::stopTimer (int destQueue, snd_seq_real_time_t real)
{
  snd_seq_event_t event;
  setOutputEventHeader (&event, real, 0, destQueue, 255, 0, 0);
  event.flags |= SND_SEQ_EVENT_LENGTH_FIXED;
  event.type = SND_SEQ_EVENT_STOP;
  
  eventOutput (&event);
}




////////////////////////////////////////////////////////////read event
int AlsaClient::getEvent (snd_seq_event_t** event)
{
  return snd_seq_event_input (handle, event);
}

int AlsaClient::drainInput ()
{
  return snd_seq_drain_output (handle);
}


////////////////////////////////////////////////////////////
namespace
{
  inline int compareTime (snd_seq_real_time_t leftTime,
			  snd_seq_real_time_t rightTime)
    {
      const int tv_sec = leftTime.tv_sec-rightTime.tv_sec;
      if (tv_sec != 0)
	return tv_sec;
      return leftTime.tv_nsec-rightTime.tv_nsec;
    }
}

void AlsaClient::sync (int destQueue, snd_seq_real_time_t real)
{
  //XXX
#if 1
# if 0
  int leftEvent = snd_seq_flush_output(handle);
  if (leftEvent < 0)
    {
      cerr << "AlsaClient::sync snd_seq_flush_output:"
	   << snd_strerror (leftEvent) << endl;
      return;
    }
# endif

  for (;;)
    {
      snd_seq_queue_info_t queueInfo;
      int tmp = snd_seq_get_queue_info (handle, destQueue, &queueInfo);
      if (tmp < 0)
	{
	  cerr << "AlsaClient::sync snd_seq_get_queue_info:"
	       << snd_strerror (tmp)
	       << endl;
	  break;
	}
      if (compareTime (real, queueInfo.time) <= 0)
	break;

# if 0
      cerr << "AlsaClient::sync "
	   << ":real.tv_sec=" << real.tv_sec
	   << ":real.tv_nsec=" << real.tv_nsec
	   << ":queueInfo.time.tv_sec=" << queueInfo.time.tv_sec
	   << ":queueInfo.tim.tv_nsec=" << queueInfo.time.tv_nsec
	   << endl;
# endif
      
      schedYield ();
# if 0
      if (leftEvent > 0)
	{
	  leftEvent = snd_seq_flush_output(handle);
	  if (leftEvent < 0)
	    {
	      cerr << "AlsaClient::sync snd_seq_flush_output:"
		   << snd_strerror (leftEvent)
		   << endl;
	      return;
	    }
	}
# endif
    }
#endif

  

#if 0
  for (;;)
    {
      snd_seq_queue_info_t queueInfo;
      int tmp = snd_seq_get_queue_info (handle, destQueue, &queueInfo);
# ifdef DEBUG
      cerr << "AlsaClient::sync left event =" << queueInfo.events << endl;
# endif
      if (tmp < 0)
	{
	  cerr << "AlsaClient::sync snd_seq_get_queue_info:"
	       << snd_strerror (tmp)
	       << endl;
	  break;
	}
      if (queueInfo.events <= 0)
	break;
    }
#endif
  
}

